## $Id: utils.py 8117 2012-04-12 06:05:29Z henrik $
##
## Copyright (C) 2011 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
"""General helper functions and utilities for the student section.
"""
import grok
from random import SystemRandom as r
from datetime import datetime
from zope.i18n import translate
from zope.component import getUtility
from reportlab.pdfgen import canvas
from reportlab.lib import colors
from reportlab.lib.units import cm
from reportlab.lib.enums import TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import (Frame, Paragraph, Image, PageBreak, Table,
                                Spacer)
from reportlab.platypus.tables import TableStyle
from reportlab.platypus.flowables import PageBreak
from zope.component import getUtility
from zope.formlib.form import setUpEditWidgets

from waeup.kofa.interfaces import IExtFileStore, IKofaUtils
from waeup.kofa.interfaces import MessageFactory as _
from waeup.kofa.students.interfaces import IStudentsUtils

SLIP_STYLE = [
    ('VALIGN',(0,0),(-1,-1),'TOP'),
    #('FONT', (0,0), (-1,-1), 'Helvetica', 11),
    ]

CONTENT_STYLE = [
    ('VALIGN',(0,0),(-1,-1),'TOP'),
    #('FONT', (0,0), (-1,-1), 'Helvetica', 8),
    #('TEXTCOLOR',(0,0),(-1,0),colors.white),
    ('BACKGROUND',(0,0),(-1,0),colors.black),
    ]

FONT_SIZE = 10
FONT_COLOR = 'black'

def formatted_label(color=FONT_COLOR, size=FONT_SIZE):
    tag1 ='<font color=%s size=%d>' % (color, size)
    return tag1 + '%s:</font>'

def trans(text, lang):
    # shortcut
    return translate(text, 'waeup.kofa', target_language=lang)

def formatted_text(text, color=FONT_COLOR, size=FONT_SIZE):
    """Turn `text`, `color` and `size` into an HTML snippet.

    The snippet is suitable for use with reportlab and generating PDFs.
    Wraps the `text` into a ``<font>`` tag with passed attributes.

    Also non-strings are converted. Raw strings are expected to be
    utf-8 encoded (usually the case for widgets etc.).

    Finally, a br tag is added if widgets contain div tags
    which are not supported by reportlab.

    The returned snippet is unicode type.
    """
    if not isinstance(text, unicode):
        if isinstance(text, basestring):
            text = text.decode('utf-8')
        else:
            text = unicode(text)
    text = text.replace('</div>', '<br /></div>')
    tag1 = u'<font color="%s" size="%d">' % (color, size)
    return tag1 + u'%s</font>' % text

def generate_student_id(students,letter):
    if letter == '?':
        letter= r().choice('ABCDEFGHKLMNPQRSTUVWXY')
    sid = u"%c%d" % (letter,r().randint(99999,1000000))
    while sid in students.keys():
        sid = u"%c%d" % (letter,r().randint(99999,1000000))
    return sid

def set_up_widgets(view, ignore_request=False):
    view.adapters = {}
    view.widgets = setUpEditWidgets(
        view.form_fields, view.prefix, view.context, view.request,
        adapters=view.adapters, for_display=True,
        ignore_request=ignore_request
        )

def render_student_data(studentview):
    """Render student table for an existing frame.
    """
    width, height = A4
    set_up_widgets(studentview, ignore_request=True)
    data_left = []
    data_right = []
    style = getSampleStyleSheet()
    img = getUtility(IExtFileStore).getFileByContext(
        studentview.context, attr='passport.jpg')
    if img is None:
        from waeup.kofa.browser import DEFAULT_PASSPORT_IMAGE_PATH
        img = open(DEFAULT_PASSPORT_IMAGE_PATH, 'rb')
    doc_img = Image(img.name, width=4*cm, height=4*cm, kind='bound')
    data_left.append([doc_img])
    #data.append([Spacer(1, 12)])
    portal_language = getUtility(IKofaUtils).PORTAL_LANGUAGE
    for widget in studentview.widgets:
        if widget.name == 'form.adm_code':
            continue
        f_label = formatted_label(size=12) % translate(
            widget.label.strip(), 'waeup.kofa',
            target_language=portal_language)
        f_label = Paragraph(f_label, style["Normal"])
        f_text = formatted_text(widget(), size=12)
        f_text = Paragraph(f_text, style["Normal"])
        data_right.append([f_label,f_text])
    table_left = Table(data_left,style=SLIP_STYLE)
    table_right = Table(data_right,style=SLIP_STYLE, colWidths=[5*cm, 6*cm])
    table = Table([[table_left, table_right],],style=SLIP_STYLE)
    return table

def render_table_data(tableheader,tabledata):
    """Render children table for an existing frame.
    """
    data = []
    #data.append([Spacer(1, 12)])
    line = []
    style = getSampleStyleSheet()
    for element in tableheader:
        field = formatted_text(element[0], color='white')
        field = Paragraph(field, style["Normal"])
        line.append(field)
    data.append(line)
    for ticket in tabledata:
        line = []
        for element in tableheader:
              field = formatted_text(getattr(ticket,element[1],u' '))
              field = Paragraph(field, style["Normal"])
              line.append(field)
        data.append(line)
    table = Table(data,colWidths=[
        element[2]*cm for element in tableheader], style=CONTENT_STYLE)
    return table

def docs_as_flowables(view, lang='en'):
    """Create reportlab flowables out of scanned docs.
    """
    # XXX: fix circular import problem
    from waeup.kofa.students.viewlets import FileManager
    from waeup.kofa.browser import DEFAULT_IMAGE_PATH
    from waeup.kofa.browser.pdf import NORMAL_STYLE, ENTRY1_STYLE
    style = getSampleStyleSheet()
    data = []

    # Collect viewlets
    fm = FileManager(view.context, view.request, view)
    fm.update()
    if fm.viewlets:
        sc_translation = trans(_('Scanned Documents'), lang)
        data.append(Paragraph(sc_translation, style["Heading3"]))
        # Insert list of scanned documents
        table_data = []
        for viewlet in fm.viewlets:
            f_label = Paragraph(trans(viewlet.label, lang), ENTRY1_STYLE)
            img_path = getattr(getUtility(IExtFileStore).getFileByContext(
                view.context, attr=viewlet.download_name), 'name', None)
            f_text = Paragraph('[Yet not provided]', ENTRY1_STYLE)
            if img_path is None:
                pass
            elif not img_path.endswith('.jpg'):
                # reportlab requires jpg images, I think.
                f_text = Paragraph('%s (Not displayable)' % (
                    viewlet.title,), ENTRY1_STYLE)
            else:
                f_text = Image(img_path, width=2*cm, height=1*cm, kind='bound')
            table_data.append([f_label, f_text])
        if table_data:
            # safety belt; empty tables lead to problems.
            data.append(Table(table_data, style=SLIP_STYLE))
    return data

def insert_footer(pdf,width,style,text=None, number_of_pages=1):
      """Render the whole footer frame.
      """
      story = []
      frame_footer = Frame(1*cm,0,width-(2*cm),1*cm)
      timestamp = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
      left_text = '<font size=10>%s</font>' % timestamp
      story.append(Paragraph(left_text, style["Normal"]))
      frame_footer.addFromList(story,pdf)
      story = []
      frame_footer = Frame(1*cm,0,width-(2*cm),1*cm)
      portal_language = getUtility(IKofaUtils).PORTAL_LANGUAGE
      right_text = translate(_('<font size=10>${a} Page ${b} of ${c}</font>',
          mapping = {'a':text, 'b':pdf.getPageNumber(), 'c':number_of_pages}),
          'waeup.kofa', target_language=portal_language)
      story.append(Paragraph(right_text, style["Right"]))
      frame_footer.addFromList(story,pdf)

class StudentsUtils(grok.GlobalUtility):
    """A collection of methods subject to customization.
    """
    grok.implements(IStudentsUtils)

    def setReturningData(self, student):
        """ This method defines what happens after school fee payment
        depending on the student's senate verdict.

        In the base configuration current level is always increased
        by 100 no matter which verdict has been assigned.
        """
        student['studycourse'].current_level += 100
        student['studycourse'].current_session += 1
        verdict = student['studycourse'].current_verdict
        student['studycourse'].current_verdict = '0'
        student['studycourse'].previous_verdict = verdict
        return

    def getPaymentDetails(self, category, student):
        """Get the payment dates of a student for the payment category
        specified.
        """
        d = {}
        d['p_item'] = u''
        d['amount'] = 0.0
        d['error'] = u''
        d['p_session'] = student['studycourse'].current_session
        session = str(d['p_session'])
        try:
            academic_session = grok.getSite()['configuration'][session]
        except KeyError:
            d['error'] = u'Session configuration object is not available.'
            return d
        d['surcharge_1'] = academic_session.surcharge_1
        d['surcharge_2'] = academic_session.surcharge_2
        d['surcharge_3'] = academic_session.surcharge_3
        if category == 'schoolfee':
            d['amount'] = academic_session.school_fee_base
            d['p_item'] = student['studycourse'].certificate.code
        elif category == 'clearance':
            d['p_item'] = student['studycourse'].certificate.code
            d['amount'] = academic_session.clearance_fee
            d['surcharge_1'] = 0.0 # no portal fee
        elif category == 'bed_allocation':
            d['p_item'] = self.getAccommodationDetails(student)['bt']
            d['amount'] = academic_session.booking_fee
            d['surcharge_1'] = 0.0 # no portal fee
        return d

    def getAccommodationDetails(self, student):
        """Determine the accommodation dates of a student.
        """
        d = {}
        d['error'] = u''
        site_configuration = grok.getSite()['configuration']
        d['booking_session'] = site_configuration.accommodation_session
        d['allowed_states'] = site_configuration.accommodation_states
        # Determine bed type
        studycourse = student['studycourse']
        certificate = getattr(studycourse,'certificate',None)
        entry_session = studycourse.entry_session
        current_level = studycourse.current_level
        if not (entry_session and current_level and certificate):
            return
        end_level = certificate.end_level
        if entry_session == grok.getSite()[
            'configuration'].accommodation_session:
            bt = 'fr'
        elif current_level >= end_level:
            bt = 'fi'
        else:
            bt = 're'
        if student.sex == 'f':
            sex = 'female'
        else:
            sex = 'male'
        special_handling = 'regular'
        d['bt'] = u'%s_%s_%s' % (special_handling,sex,bt)
        return d

    def selectBed(self, available_beds):
        """Select a bed from a list of available beds.

        In the base configuration we select the first bed found,
        but can also randomize the selection if we like.
        """
        return available_beds[0]

    def renderPDF(self, view, filename='slip.pdf',
        student=None, studentview=None, tableheader=None, tabledata=None):
        """Render pdf slips for various pages.
        """
        # XXX: we have to fix the import problems here.
        from waeup.kofa.browser.interfaces import IPDFCreator
        from waeup.kofa.browser.pdf import NORMAL_STYLE, ENTRY1_STYLE
        style = getSampleStyleSheet()
        creator = getUtility(IPDFCreator)
        data = []
        doc_title = view.label
        author = '%s (%s)' % (view.request.principal.title,
                              view.request.principal.id)
        footer_text = view.label
        if getattr(student, 'student_id', None) is not None:
            footer_text = "%s - %s - " % (student.student_id, footer_text)

        # Insert student data table
        portal_language = getUtility(IKofaUtils).PORTAL_LANGUAGE
        if student is not None:
            bd_translation = trans(_('Base Data'), portal_language)
            data.append(Paragraph(bd_translation, style["Heading3"]))
            data.append(render_student_data(studentview))

        # Insert widgets
        data.append(Paragraph(view.title, style["Heading3"]))
        portal_language = getUtility(IKofaUtils).PORTAL_LANGUAGE
        table = creator.getWidgetsTable(
            view.form_fields, view.context, None, lang=portal_language)
        data.append(table)

        # Insert scanned docs
        data.extend(docs_as_flowables(view, portal_language))

        # Insert content table on second page
        if tabledata and tableheader:
            data.append(PageBreak())
            data.append(Paragraph(view.content_title, style["Heading3"]))
            contenttable = render_table_data(tableheader,tabledata)
            data.append(contenttable)

        view.response.setHeader(
            'Content-Type', 'application/pdf')
        try:
            pdf_stream = creator.create_pdf(
                data, None, doc_title, author=author, footer=footer_text)
        except IOError:
            view.flash('Error in image file.')
            return view.redirect(view.url(view.context))
        return pdf_stream

    VERDICTS_DICT = {
        '0': _('(not yet)'),
        'A': 'Successful student',
        'B': 'Student with carryover courses',
        'C': 'Student on probation',
        }

    SEPARATORS_DICT = {
        }
